/*         Snakman v0.2
	Hybrid of Snake and Pacman
*/

/* 
	Credits:
		Thanks to Shine for the Original Blitting Functions
		function stubs: PSPDev - Browser Api (by djhuevo & neofar, http://pspdev.ofcode.com/api.php)
		controller function usage based on a program by skippy911
*/

#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspctrl.h>
#include <stdlib.h>
#include <string.h>
#include <psppower.h>
#include <math.h>

/* Define the module info section */
PSP_MODULE_INFO("Snakman v0.2", 0, 1, 1);

/* Define the main thread's attribute value (optional) */
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER | THREAD_ATTR_VFPU);

//External Files (Images)
extern u32 background_start[];
extern u32 headR_start[];
extern u32 headT_start[];
extern u32 headL_start[];
extern u32 headB_start[];
extern u32 headRC_start[];
extern u32 headTC_start[];
extern u32 headLC_start[];
extern u32 headBC_start[];
extern u32 apple_start[];
extern u32 orange_start[];
extern u32 red_start[];
extern u32 pause_start[];

//Arrays of Images
u32* heads[] = { headR_start, headT_start, headL_start, headB_start, headRC_start, headTC_start, headLC_start, headBC_start };
u32* badGuys[] = { red_start, orange_start };

//Define Edges
#define RIGHT  0
#define TOP    1
#define LEFT   2
#define BOTTOM 3

/* game structs and variables */
//Define What A Cell Is (Last TBW Reminent -- used for the "apples")
typedef struct
{
	int x;
	int y;
	int direction;
} Cell;

//Define A Point (Again, used for the "apples")
typedef struct {
	int x;
	int y;
} point;

//Color "Functions" (Conversion and Test)
#define RGB(r, g, b) ((r)|((g)<<8)|((b)<<16))
#define IS_ALPHA(value) (((value)&0xff000000)==0)

//Set Text Color (Scores on Right Hand Side)
#define TEXT_COLOR RGB(204, 204, 204)

Cell target;		//Where the "Apple" Is
u32* targetImage;	//Which Image to use for the "Apple"
int charX=16;		//Initialize the Pacman Guy's Position
int charY=16;		//To (16, 16)
int score;
int high = 0;		//Declare and Initialize High to Zero
					//Reinitialized in newGame();

int speed = 3;		//Define Speed of the Pacman Guy

//Define Traversable Area for Pacman Guy
int minX = 16;		//Left Limit
int minY = 16;		//Lower Limit
int maxX = 352;		//Right Limit
int maxY = 256;		//Top Limit

//Initalize The Enemy Positions
//Plus 16 because that's the upper Left Hand Corner - (16, 16)
int redX    = 40 +16;
int redY    = 0  +16;
int orangeX = 0  +16;
int orangeY = 40 +16;

//Initialize The Enemy Speeds
int orangeVelX = 2;
int orangeVelY = 2;
int redVelX    = 3;
int redVelY    = 3;

//Initialize the Pacman Guy's Look,
//Closed Mouth, Pointing Right
int mouth      = 0;
int dir        = 0;

//Initialize Random Number Seeds for the "Apple"
static unsigned seed1;    
static unsigned seed2; 

//constants
#define SCREEN_WIDTH 480
#define SCREEN_HEIGHT 272

#define	PSP_PIXEL_FORMAT 3  // 32 bit RGBA
#define	PSP_LINE_SIZE 512   // in u32

//480*272 = 60*38
#define CMAX_X 60
#define CMAX_Y 38
#define CMAX2_X 30
#define CMAX2_Y 19
#define CMAX4_X 15
#define CMAX4_Y 9

//For _ltoa
#define MAX_LENGTH 32+1

//variables
u32* pg_vramtop = (u32*)(0x04000000+0x40000000);
long pg_screenmode;
long pg_showframe;
long pg_drawframe;

void pgWaitVn(unsigned long count) {
	for (; count>0; --count) {
		sceDisplayWaitVblankStart();
	}
}

void pgWaitV() {
	sceDisplayWaitVblankStart();
}

u32* pgGetVramAddr(unsigned long x,unsigned long y) {
	return pg_vramtop+(y * PSP_LINE_SIZE)+x+(pg_drawframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0);
}

void pgScreenFrame(long mode,long frame) {
	pg_screenmode=mode;
	frame=(frame?1:0);
	pg_showframe=frame;
	if (mode==0) {
		//screen off
		pg_drawframe=frame;
		sceDisplaySetFrameBuf(0,0,0,1);
	} else if (mode==1) {
		//show/draw same
		pg_drawframe=frame;
		sceDisplaySetFrameBuf((char*)(pg_vramtop+(pg_drawframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0)),PSP_LINE_SIZE,PSP_PIXEL_FORMAT,1);
	} else if (mode==2) {
		//show/draw different
		pg_drawframe=(frame?0:1);
		sceDisplaySetFrameBuf((char*)(pg_vramtop+(pg_drawframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0)),PSP_LINE_SIZE,PSP_PIXEL_FORMAT,1);
	}
}

void pgInit() {
	sceDisplaySetMode(0,SCREEN_WIDTH,SCREEN_HEIGHT);
	pgScreenFrame(0,0);
}

void pgScreenFlip() {
	pg_showframe=(pg_showframe?0:1);
	pg_drawframe=(pg_drawframe?0:1);
	sceDisplaySetFrameBuf((char*)(pg_vramtop+(pg_showframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0)),PSP_LINE_SIZE,PSP_PIXEL_FORMAT,0);
}

void pgScreenFlipV() {
	pgWaitV();
	pgScreenFlip();
}

void pgFlipDrawFrame() {
	pg_drawframe=(pg_drawframe?0:1);
}

void pgFlipShowFrameV() {
	pgWaitV();
	pg_showframe=(pg_showframe?0:1);
	sceDisplaySetFrameBuf((char*)(pg_vramtop+(pg_showframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0)),PSP_LINE_SIZE,PSP_PIXEL_FORMAT,0);
}

/* Exit callback */
int exit_callback(void) {
	sceKernelExitGame();

	return 0;
}

/* Callback thread */
void CallbackThread(void *arg) {
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void) {
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0) {
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

void initSeeds()  {
	//For "Apple" Random Placement, Set Seeds to Current Time
	seed1 = sceKernelLibcTime((void *) 0); 
	seed2 = sceKernelLibcTime((void *) 0); 
} 

unsigned int myRand() {
	//Gengerate Random Number
	seed1 = (seed1 + 46906481) ^ seed2; 
	seed2 = seed1 ^ ( ((seed2<<3) | (seed2 >> 29)) + 103995407); 
    
	return seed1 - seed2; 
}

//Set Up Font For the Score Display
//(Only contains 0-9)
//"1" Gets Written to with the Color of the Text, "0" is left Alone
const char digitFont[] = {
	1,1,1,1,
	1,0,0,1,
	1,0,0,1,
	1,0,0,1,
	1,1,1,1,
	
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	
	1,1,1,1,
	0,0,0,1,
	1,1,1,1,
	1,0,0,0,
	1,1,1,1,
	
	1,1,1,1,
	0,0,0,1,
	1,1,1,1,
	0,0,0,1,
	1,1,1,1,
	
	1,0,0,1,
	1,0,0,1,
	1,1,1,1,
	0,0,0,1,
	0,0,0,1,
	
	1,1,1,1,
	1,0,0,0,
	1,1,1,1,
	0,0,0,1,
	1,1,1,1,
	
	1,1,1,1,
	1,0,0,0,
	1,1,1,1,
	1,0,0,1,
	1,1,1,1,
	
	1,1,1,1,
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	
	1,1,1,1,
	1,0,0,1,
	1,1,1,1,
	1,0,0,1,
	1,1,1,1,

	1,1,1,1,
	1,0,0,1,
	1,1,1,1,
	0,0,0,1,
	1,1,1,1
};

//Print out A Digit
void print7Segment(int x, int y, int digit) {
	u32* vram = pgGetVramAddr(x, y);
	int xo, yo;
	int i = digit * 5*4;
	for (yo = 0; yo < 5; yo++) {
		for (xo = 0; xo < 4; xo++) {
			if (digitFont[i]) {
				vram[xo + yo * PSP_LINE_SIZE] = TEXT_COLOR;
			} else {
				vram[xo + yo * PSP_LINE_SIZE] = background_start[x + xo + (y + yo) * SCREEN_WIDTH];
			}
			i++;
		}
	}
}

//Print Out A Number
void printDecimal(int x, int y, int color, int value) {
	int i;
	int digits[6];
	for (i = 5; i >= 0; i--) {
		digits[i] = value%10;
		value /= 10;
	}
	i = 0;
	while (i < 5 && digits[i] == 0) digits[i++] = -1;
	for (i = 0; i < 6; i++) {
		if (digits[i] >= 0) {
			print7Segment(x, y, digits[i]);
			x += 6;
		}
	}
}

//Erase a Cell (Where the "Apple" was Previously)
void fillCell(int x, int y) {
	u32* vram = pgGetVramAddr(x * 16, y * 16);
	int xo, yo;
	for (yo = 0; yo < 16; yo++) {
		for (xo = 0; xo < 16; xo++) {
			vram[xo + yo * PSP_LINE_SIZE] = background_start[16*x + xo + (16*y + yo) * SCREEN_WIDTH];
		}
	}
}

//Put the "Apple" Somewhere
void plotCellImage(int x, int y, u32* image) {
	u32* vram = pgGetVramAddr(x * 16, y * 16);
	int xo, yo;
	for (yo = 0; yo < 16; yo++) {
		for (xo = 0; xo < 16; xo++) {
			if (!IS_ALPHA(*image)) {
				vram[xo + yo * PSP_LINE_SIZE] = *image;
			} else vram[xo + yo * PSP_LINE_SIZE] = background_start[16*x + xo + (16*y + yo) * SCREEN_WIDTH];
			image++;
		}
	}
}

//Put an Image somewhere (non TBW)
void plotImage(int x, int y, u32* image, int sizeX, int sizeY) {
	u32* vram = pgGetVramAddr(x, y);
	int xo, yo;
	for (yo = 0; yo < sizeY; yo++) {
		for (xo = 0; xo < sizeX; xo++) {
			if (!IS_ALPHA(*image)) {
				vram[xo + yo * PSP_LINE_SIZE] = *image;
			} else {
				vram[xo + yo * PSP_LINE_SIZE] = background_start[x + xo + (y + yo) * SCREEN_WIDTH];  //This line wasn't as stupid as I thought.
			}
			image++;
		}
	}
}

//Better Plotting for Overlapping Images (can cause problems other times, so ^^ is used most of the time)
void plotImageLayers(int x, int y, u32* image, int sizeX, int sizeY) {
	u32* vram = pgGetVramAddr(x, y);
	int xo, yo;
	for (yo = 0; yo < sizeY; yo++) {
		for (xo = 0; xo < sizeX; xo++) {
			if (!IS_ALPHA(*image)) {
				vram[xo + yo * PSP_LINE_SIZE] = *image;
			} else {
				//vram[xo + yo * PSP_LINE_SIZE] = background_start[x + xo + (y + yo) * SCREEN_WIDTH];  //But I need it out...  sometimes...
			}
			image++;
		}
	}
}

//Generate A Random x, y and put the "Apple" in that Cell
void createRandomTarget() {
	target.x = (myRand() % 21) + 1;
	target.y = (myRand() % 15) + 1;
	targetImage = apple_start;
}

//Screenshot Code
/*
void writeByte(int fd, unsigned char data) {
	sceIoWrite(fd, &data, 1);
}


void screenshot()
{
	const char tgaHeader[] = {0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	const int width = SCREEN_WIDTH;
	const int lineWidth = PSP_LINE_SIZE;
	const int height = SCREEN_HEIGHT;
	unsigned char lineBuffer[width*4];
	u32* vram = pgGetVramAddr(0, 0);
	int x, y;
	int fd = sceIoOpen("ms0:/snake-screenshot.tga", PSP_O_CREAT | PSP_O_TRUNC | PSP_O_WRONLY, 0777);
	if (!fd) return;
	sceIoWrite(fd, tgaHeader, sizeof(tgaHeader));
	writeByte(fd, width & 0xff);
	writeByte(fd, width >> 8);
	writeByte(fd, height & 0xff);
	writeByte(fd, height >> 8);
	writeByte(fd, 24);
	writeByte(fd, 0);
	for (y = height - 1; y >= 0; y--) {
		for (x = 0; x < width; x++) {
			u32 color = vram[y * lineWidth + x];
			unsigned char red = color & 0xff;
			unsigned char green = (color >> 8) & 0xff;
			unsigned char blue = (color >> 16) & 0xff;
			lineBuffer[3*x] = blue;
			lineBuffer[3*x+1] = green;
			lineBuffer[3*x+2] = red;
		}
		sceIoWrite(fd, lineBuffer, width * 3);
	}
	sceIoClose(fd);
}*/

//Integer to String Conversion
char* _ltoa(char* buf, long i, int base) {
    char reverse[MAX_LENGTH+1];   // plus one for the null
    char* s;
    char sign = i < 0;

    i = labs(i);
    reverse[MAX_LENGTH] = 0;
    s = reverse+MAX_LENGTH;
    do {
        *--s = "0123456789abcdefghijklmnopqrstuvwxyz"[i % base];
        i /= base;
    } while (i);
    if (sign) *--s = '-';
    return strcpy(buf, s);
}

//String to Integer Conversion
int stoi(char num) {
	// 0<=num<=9
	switch(num) {
		case '0':
			return 0;
			break;
		case '1':
			return 1;
			break;
		case '2':
			return 2;
			break;
		case '3':
			return 3;
			break;
		case '4':
			return 4;
			break;
		case '5':
			return 5;
			break;
		case '6':
			return 6;
			break;
		case '7':
			return 7;
			break;
		case '8':
			return 8;
			break;
		case '9':
			return 9;
			break;
		default:
			return -1;
			break;
	}
}

//Write Score to File
void saveScore(int theScore) {
	int fd = sceIoOpen("ms0:/snakman.dat", PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);
	
	char scoreString[9];
	_ltoa(scoreString, theScore, 10);
	
	//char scoreString[] = itoa(theScore);
	int i;
	for(i=0; i<9; i++) {
		if(stoi(scoreString[i]) == -1) {
			break;
		}
	}
	for(i=i; i<9; i++) {
		scoreString[i] = ' ';
	}
	
	sceIoWrite(fd, scoreString, 9);
	sceIoClose(fd);
}

//Read Score from File
int readScore() {
	int fd = sceIoOpen("ms0:/snakman.dat", PSP_O_RDONLY, 0777);
	if(fd < 0) {
		saveScore(0);
		return 0;
	}
	
	//char scoreString[] = itoa(theScore);
	char scoreString[11];
	
	sceIoRead(fd, scoreString, 9);
	sceIoClose(fd);

	int x = 0;
	int i;
	for(i=0; scoreString[i] != ' '; i++)
	{
		x *= 10;
		x += stoi (scoreString[i]);
	}
	return x;
}

//Reset all the Variables
void newGame()
{
	pgFlipDrawFrame();
	u32* vram = pgGetVramAddr(0, 0);
	int xo, yo;
	u32* image = background_start;
	for (yo = 0; yo < SCREEN_HEIGHT; yo++) {
		for (xo = 0; xo < SCREEN_WIDTH; xo++) {
			if (!IS_ALPHA(*image)) vram[xo + yo * PSP_LINE_SIZE] = (*image)& 0xffffff;
			image++;
		}
	}
	
	high = readScore();
	charX = 16;
	charY = 16;
	createRandomTarget();

	minX = 16;
	minY = 16;
	maxX = 352;
	maxY = 256;

	redX    = 40 +16;
	redY    = 0  +16;
	orangeX = 0  +16;
	orangeY = 40 +16;

	orangeVelX = 2;
	orangeVelY = 2;
	redVelX    = 3;
	redVelY    = 3;

	mouth = 0;

	score = 0;

	pgFlipShowFrameV();

	plotCellImage(target.x, target.y, targetImage);
	plotImage(charX, charY, heads[0], 32, 32);
}

//Do A Frame of the Enemies
void balls() {
	//orangeVelX orangeVelY redVelX redVelY redX redY orangeX orangeY badGuys[red, orange]
	u32* vram  = pgGetVramAddr(orangeX, orangeY);
	u32* vram2 = pgGetVramAddr(redX, redY);
	int xo, yo;
	for (yo = 0; yo < 32; yo++) {
		for (xo = 0; xo < 32; xo++) {
			vram[xo + yo * PSP_LINE_SIZE] = background_start[1*orangeX + xo + (1*orangeY + yo) * SCREEN_WIDTH];
			vram2[xo + yo * PSP_LINE_SIZE] = background_start[1*redX + xo + (1*redY + yo) * SCREEN_WIDTH];
		}
	}

	orangeX += orangeVelX;
	orangeY += orangeVelY;
	redX    += redVelX;
	redY    += redVelY;

	if(orangeY <= minY) {
		orangeY = minY;
		orangeVelY *= -1;
	}
	if(orangeY >= maxY-32) {
		orangeY = maxY-33;
		orangeVelY *= -1;
	}
	if(orangeX <= minX) {
		orangeX = minX;
		orangeVelX *= -1;
	}
	if(orangeX >= maxX-32) {
		orangeX = maxX-33;
		orangeVelX *= -1;
	}

	if(redY <= minY) {
		redY = minY;
		redVelY *= -1;
	}
	if(redY >= maxY-32) {
		redY = maxY-33;
		redVelY *= -1;
	}
	if(redX <= minX) {
		redX = minX;
		redVelX *= -1;
	}
	if(redX >= maxX-32) {
		redX = maxX-33;
		redVelX *= -1;
	}

	plotImageLayers(orangeX, orangeY, badGuys[0], 32, 32);
	plotImageLayers(redX, redY, badGuys[1], 32, 32);
}

//Gets the Distance between two Circles' radii (Is that the plural of "radius?")
float circleDistance(point center1, point center2) {
	float distance;
	
	distance = (center1.x-center2.x)*(center1.x-center2.x)+(center1.y-center2.y)*(center1.y-center2.y);
	distance = sqrt(distance);

	return distance;
}

//See if Pacman Guy Has Hit a Ball
int checkDead() {
	point charCenter;
	point redCenter;
	point orangeCenter;

	charCenter.x = charX + 16;
	charCenter.y = charY + 16;

	redCenter.x = redX + 16;
	redCenter.y = redY + 16;

	orangeCenter.x = orangeX + 16;
	orangeCenter.y = orangeY + 16;

	if(circleDistance(charCenter, redCenter) < 28) {
		return 1;
	}
	if(circleDistance(charCenter, orangeCenter) < 28) {
		return 1;
	}
	return 0;
}

//Get Button Input and Move Pacman Guy if Necessary
void move()
{
	static int changed = 0;
	changed = 0;
	int xo, yo;

	SceCtrlData pad;
	sceDisplayWaitVblankStart();
	sceCtrlReadBufferPositive(&pad, 1); 

	int xOld = charX;
	int yOld = charY;

	if (pad.Buttons & PSP_CTRL_UP) {
		changed = 1;
		dir = 1;
		if(charY >= minY+speed) {
			charY -= speed;
		}
	} else if (pad.Buttons & PSP_CTRL_DOWN) { 
		changed = 1;
		dir = 3;
		if(charY <= maxY-speed-32) {
			charY += speed;
		}
	} else if (pad.Buttons & PSP_CTRL_LEFT) { 
		changed = 1;
		dir = 2;
		if(charX >= minX+speed) {
			charX -= speed;
		}
	} else if (pad.Buttons & PSP_CTRL_RIGHT) { 
		changed = 1;
		dir = 0;
		if(charX <= maxX-speed-32) {
			charX += speed;
		}
	} else if (pad.Buttons & PSP_CTRL_START) {
		plotImage(138, 115, pause_start, 89, 15);
		pgWaitVn(25);
		while(1) {
			SceCtrlData pad2;
			sceDisplayWaitVblankStart();
			sceCtrlReadBufferPositive(&pad2, 1); 

			if (pad2.Buttons & PSP_CTRL_START) {
				u32* vram = pgGetVramAddr(138, 115);
				for (yo = 0; yo < 138; yo++) {
					for (xo = 0; xo < 115; xo++) {
						vram[xo + yo * PSP_LINE_SIZE] = background_start[1*138 + xo + (1*115 + yo) * SCREEN_WIDTH];
					}
				}
				break;
			}
		}
		plotCellImage(target.x, target.y, targetImage);
		plotImage(orangeX, orangeY, badGuys[0], 32, 32);
		plotImage(redX, redY, badGuys[1], 32, 32);
		plotImage(charX, charY, heads[dir+mouth], 32, 32);
		pgWaitVn(10);
	}

	
	if(changed == 1) {
		u32* vram = pgGetVramAddr(xOld, yOld);
		for (yo = 0; yo < 32; yo++) {
			for (xo = 0; xo < 32; xo++) {
				vram[xo + yo * PSP_LINE_SIZE] = background_start[1*xOld + xo + (1*yOld + yo) * SCREEN_WIDTH];
			}
		}
		
		if((charX <= (target.x*16)+14 && charX >= (target.x*16)-30) && (charY <= (target.y*16)+14 && charY >= (target.y*16)-30)) {
			fillCell(target.x, target.y);
			createRandomTarget();
			score++;
		}
	}
	plotCellImage(target.x, target.y, targetImage);
	plotImage(charX, charY, heads[dir+mouth], 32, 32);
}
	
int main(void)
{
	//scePowerSetClockFrequency(333, 333, 166);

	initSeeds();
	SetupCallbacks();

	pgInit();
	pgScreenFrame(1, 1);
	
	newGame();

	int i = 0;
	while (1) {
		//for (i = 0; i < 5; i++) keyboardControl();
		move();
		balls();
		checkDead();

		if (score > high) {
			high = score;
		}

		if (checkDead()) {
			plotCellImage(target.x, target.y, targetImage);
			plotImageLayers(orangeX, orangeY, badGuys[0], 32, 32);
			plotImageLayers(redX, redY, badGuys[1], 32, 32);
			plotImageLayers(charX, charY, heads[dir+mouth], 32, 32);

			if(score == high) {
				saveScore(score);
			}

			pgWaitVn(50);
			newGame();
		}
		printDecimal(425, 70, TEXT_COLOR, score);
		printDecimal(425, 89, TEXT_COLOR, high);
		if(i%10 == 0) {
			if(mouth == 0) {
				mouth = 4;
			} else if(mouth == 4) {
				mouth = 0;
			}
			i = 0;
		}
		i++;
	}
	return 0;
}
